package wacc

import wacc.ast.Position
import wacc.types._

enum Error {
  case DuplicateDeclaration(ident: ast.Ident)
  case UndefinedIdentifier(ident: ast.Ident, identType: renamer.IdentType)

  case FunctionParamsMismatch(pos: Position, expected: Int, got: Int)
  case SemanticError(pos: Position, msg: String)
  case TypeMismatch(pos: Position, expected: SemType, got: SemType, msg: String)
  case InternalError(pos: Position, msg: String)
}

def printError(error: Error)(using errorContent: String): Unit = {
  println("Semantic error:")
  error match {
    case Error.DuplicateDeclaration(ident) =>
      printPosition(ident.getPosition)
      println(
        s"Duplicate declaration of identifier ${ident.v}"
      )
      highlight(ident.getPosition, ident.v.length)
    case Error.UndefinedIdentifier(ident, identType) =>
      printPosition(ident.getPosition)
      println(
        s"Undefined ${identType.toString.toLowerCase()} ${ident.v}"
      )
      highlight(ident.getPosition, ident.v.length)
    case Error.FunctionParamsMismatch(pos, expected, got) =>
      printPosition(pos)
      println(s"Function expects $expected parameters, got $got")
      highlight(pos, 1)
    case Error.TypeMismatch(pos, expected, got, msg) =>
      println(s"Type mismatch: expected $expected, got $got")
    case Error.SemanticError(pos, msg) =>
      printPosition(pos)
      println(s"$msg at line: ${pos.line} column: ${pos.column}")
      highlight(pos, 1)
    case wacc.Error.InternalError(pos, msg) =>
      printPosition(pos)
      println(s"Internal error: $msg")
      highlight(pos, 1)
  }

}

def highlight(pos: Position, size: Int)(using errorContent: String): Unit = {
  val lines = errorContent.split("\n")

  val preLine = if (pos.line > 1) lines(pos.line - 2) else ""
  val midLine = lines(pos.line - 1)
  val postLine = if (pos.line < lines.size) lines(pos.line) else ""
  val linePointer = " " * (pos.column + 2) + ("^" * (size)) + "\n"

  println(
    s"  >$preLine\n  >$midLine\n$linePointer  >$postLine"
  )
}

def printPosition(pos: Position): Unit = {
  println(s"(line ${pos.line}, column ${pos.column}):")
}
